!**********************************************************************!
!----------------------------------------------------------------------!
!--------------------------- 1D Heat Conduction -----------------------!
!-------------------------- With Change of Phase ----------------------!
!----------------------------------------------------------------------!
!----------------------------------- By -------------------------------!
!----------------------------------------------------------------------!
!------------------------------ Olav Rømcke ---------------------------!
!----------------------------------------------------------------------!
!------------------------------- May 2018 -----------------------------!
!----------------------------------------------------------------------!
!**********************************************************************!



!Physical variables
MODULE FIELD
IMPLICIT NONE
SAVE

INTEGER,PARAMETER::NINP=21,ANIMATION=22, RUNDATA=23, INTERP=24
INTEGER::IBAR,IM1,IMAX,I,J,NFRAME,NTHFRAME,&
UCOUNT,UCOUNT_T,FLAG,PREFNODE,ENTRIESA_T
DOUBLE PRECISION::DELT,T,TMAX,DELX,XMAX,RDX,TH,TC,TIN,TL,TS,STE,VF,&
TH_RE,TC_RE,TI_RE,TL_RE,TS_RE,THERM_EXP,T_MID,FSLIM,CPU_START,CPU_STOP,DELTN,FSE,PX
DOUBLE PRECISION,ALLOCATABLE,DIMENSION(:)::X,TEMP,TEMP_P,FS,&
FS_P,FS_N,XI,XI_N,TB,TAVAL,TEMP_VEC
INTEGER, ALLOCATABLE, DIMENSION (:) :: NNT,TI,TJ

END MODULE FIELD

!Solver variables
MODULE SOLVEFIELD
IMPLICIT NONE
SAVE

INTEGER :: ITER_T,ITER_TT,ITMAX_T,ITMAX_TT
DOUBLE PRECISION :: RES_T,RES_TT,EPSI_T,EPSI_TT

END MODULE SOLVEFIELD


!------------------------------------------------------------------
FUNCTION F_INV(x,xe,ys,yl)			
! Inverse of solid fraction temperature relation for substance	
! Here: linear eutectic	
! x = Solid fraction, (e-eutectic)
! y = Melting temperature (s-solidus, l-liquidus)

  DOUBLE PRECISION :: F_INV,x,xe,ys,yl
  
  ! linear eutectic
  F_INV = max(yl-x/xe*(yl-ys),ys)


RETURN
END FUNCTION F_INV
!-------------------------------------------------------------------








PROGRAM BAR1D
!-------------------------------------------------------------
!     SOLVES ENERGY EQUATION WITH A CHANGE IN PHASE
!     Version 01.May 2018 Olav Rømcke
!------------------------------------------------------------
	



USE FIELD
USE SOLVEFIELD

IMPLICIT NONE




CALL INPUT                ! Read input
CALL PREPARE_TECPLOT_FILE ! Prepare files for writing data
CALL INIT_T  		  ! Define initial Thermal fields.
CALL BCTEMP   		  ! Impose boundary conditions by calling BCTEMP.
CALL NODE_NUMBERS_TEMP	  ! define node numbers for temperature field
CALL GEN_T_MAT            ! Generate system matrix





! Capture physics every nth timestep
NTHFRAME = CEILING(DELTN/DELT)
NFRAME = 0

! start CPU-clock
CALL CPU_TIME(CPU_START)


!     ==== START OF GLOBAL LOOP ====
  
DO WHILE(.TRUE.)
  ! Update time
  T=T+DELT
  
  ! Calculate Thermal fields
  CALL TITER

  ! Capture Physics
  IF(MOD(NFRAME,NTHFRAME).EQ.0) THEN
    CALL WRITE_TO_TECPLOT
  ENDIF
  NFRAME = NFRAME + 1
  CALL WRITE_TO_TECPLOT2

  ! Data to screen
  WRITE(*,*)
  WRITE(*,*)
  WRITE(*,'(A,G12.5)') ' Time = ',T
  WRITE(*,*)'-------------------------------------------------------------------------------------'
  WRITE(*,'(2(A,I12))')  '  TT-iter = ',ITER_TT,' |  T-iter = ',ITER_T
  WRITE(*,'(2(A,G12.5))')'  TT-res  = ',RES_TT, ' |  T-res  = ',RES_T
  WRITE(*,'(2(A,G12.5))')'  FS-glob = ',VF,     ' |  PP-pos = ',PX
  WRITE(*,*)'-------------------------------------------------------------------------------------'

  IF(T.GE.TMAX)EXIT
END DO

!     ==== END OF GLOBAL LOOP ====


! stop CPU-clock
CALL CPU_TIME(CPU_STOP)










! End routine
WRITE(*,*)'Simulation finished.'
WRITE(*,*)'Duration: ', CPU_STOP-CPU_START, ' sec'


! Deallocate
DEALLOCATE (TEMP,TEMP_P,X,FS,FS_P,&
NNT,TAVAL,TI,TJ,TB,TEMP_VEC)

STOP
END PROGRAM

!------------------------------------------------------------

SUBROUTINE INPUT
!     To read input data and compute parameters from input file 
USE FIELD
USE SOLVEFIELD
IMPLICIT NONE
INTEGER ::IFAIL
DOUBLE PRECISION ::DUMMY1,DUMMY2

OPEN (UNIT=NINP,FILE='SIMINP1D.INP',FORM='FORMATTED')
READ(NINP,1000)
READ(NINP,1020)TMAX,DELT,DELTN,DELX,XMAX,STE,TH_RE,TC_RE,TI_RE,TL_RE,TS_RE,&
FSE,DUMMY1,DUMMY2,EPSI_T,EPSI_TT



!Define Grid
IBAR=NINT(DELX)
DELX=XMAX/DELX
IM1=IBAR+1
IMAX=IBAR+2
RDX  = 1.0/DELX

! Stop criterions
ITMAX_T=NINT(DUMMY1)
ITMAX_TT=NINT(DUMMY2)

! Set time to zero initially
T=0.0

! Solid liquid interphase limit
FSLIM=FSE

! Temperature scale
TH=(TH_RE-TC_RE)/(TH_RE-TC_RE)
TC=(TC_RE-TC_RE)/(TH_RE-TC_RE)
TIN=(TI_RE-TC_RE)/(TH_RE-TC_RE)
TL=(TL_RE-TC_RE)/(TH_RE-TC_RE)
TS=(TS_RE-TC_RE)/(TH_RE-TC_RE)

! entries TA (temperature matrix)
	    ! in domain	  ! walls
ENTRIESA_T = 3*(IBAR-2) + 2*2

ALLOCATE (TEMP(IBAR+2),TEMP_P(IBAR+2),X(IBAR+2),FS(IBAR+2),&
FS_P(IBAR+2),FS_N(IBAR+2),XI(IBAR+2),XI_N(IBAR+2),&
NNT(IBAR+2),TAVAL(ENTRIESA_T),TI(ENTRIESA_T),&
TJ(ENTRIESA_T),TB(IBAR),TEMP_VEC(IBAR))

1000 FORMAT(//////)
1020 FORMAT(10(43X,F20.0/))
RETURN
END 

!------------------------------------------------------------

SUBROUTINE PREPARE_TECPLOT_FILE
! Prepare header in .DAT files for logging results
USE FIELD
IMPLICIT NONE


DO I = 2,IM1
    X(I) = (I-1.5)*DELX
ENDDO

OPEN (UNIT=ANIMATION,FILE='TECPLOT.DAT',FORM='FORMATTED')  ! Output file
WRITE(ANIMATION,*)'TITLE     = "Animate"'
WRITE(ANIMATION,*)'VARIABLES = "X","T","FS"'
WRITE(ANIMATION,*)''

OPEN (UNIT=RUNDATA,FILE='TECPLOT2.DAT',FORM='FORMATTED')  ! Output file
WRITE(RUNDATA,*)'TITLE     = "Sim Data"'
WRITE(RUNDATA,*)'VARIABLES = "T", "PX", "VF", "RES_T", "ITER_T", "RES_TT", "ITER_TT"'
WRITE(RUNDATA,*)''

RETURN
END 

!------------------------------------------------------------

SUBROUTINE WRITE_TO_TECPLOT
!     Logging physical data for use in Techplot
USE FIELD
IMPLICIT NONE


! Write to ANIMATION file
WRITE(ANIMATION,*)'ZONE ','T="time = ',T,'" I=',(IM1-1),', DATAPACKING=POINT'
  DO I=2,IM1
    WRITE(ANIMATION,*)X(I),' ',TEMP(I),' ',FS(I)
ENDDO
WRITE(ANIMATION,*)


RETURN
END 

!------------------------------------------------------------

SUBROUTINE WRITE_TO_TECPLOT2
!     Logging simulation (meta) data for use in Tecplot
USE FIELD
USE SOLVEFIELD
IMPLICIT NONE

!Calculate Global solid fraction
VF=0.0

DO I=2,IM1
    VF = VF + FS(I)*DELX/(XMAX)
ENDDO


! Calculate solid-liquid Phase front x coordinates
PX=0.0

DO I=2,IM1
   IF ((FS(I).GT.(FSLIM)).AND.(FS(I+1).LE.(FSLIM))) THEN
     PX=X(I) + DELX*(FSLIM-FS(I))/(FS(I+1)-FS(I))
   ENDIF
   IF ((FS(I).LE.(FSLIM)).AND.(FS(I+1).GT.(FSLIM))) THEN
     PX=X(I) - DELX*(FS(I+1)-FSLIM)/(FS(I+1)-FS(I))
   ENDIF
ENDDO


! Write to run data file
WRITE(RUNDATA,*)T,' ',PX,' ',VF,' ',RES_T,' ',ITER_T,' ',RES_TT,' ',ITER_TT



RETURN
END 

!------------------------------------------------------------


SUBROUTINE INIT_T
!     Purpose:
!     To initiate the Thermal and Momentum fields 
USE FIELD
IMPLICIT NONE

DO I=1,IMAX

    TEMP(I)=TIN
    FS(I)=1.0

ENDDO 
RETURN
END

!------------------------------------------------------------

SUBROUTINE NODE_NUMBERS_TEMP
!     purpose:
!     To assign node numbers to thermal fields
!     UCOUNT_T = number of unknowns
!     NNT(i) = node number associated with node at position (i)
USE FIELD
IMPLICIT NONE

NNT(:) = 0
UCOUNT_T = 0		

DO I=2,IM1
  UCOUNT_T = UCOUNT_T + 1
  NNT(I) = UCOUNT_T
ENDDO


RETURN
END

!------------------------------------------------------------


SUBROUTINE BCTEMP
!     purpose:
!     To give boundary conditions to the temp and fs
USE FIELD
IMPLICIT NONE


  TEMP(1) = TH;
  TEMP(IMAX) = TC;
  FS(1)=FS(2);
  FS(IMAX)=FS(IMAX-1);



RETURN
END

!------------------------------------------------------------


SUBROUTINE GEN_T_MAT 
!     purpose:
!     To construct the sparse coefficient matrix A_T for solving temperature
USE FIELD
IMPLICIT NONE
INTEGER :: N
DOUBLE PRECISION :: APO,AP,AE,AW


      
  N = 0

  DO I=2,IM1
    APO = 1.0
    AE = -DELT/DELX**2
    AW = -DELT/DELX**2

    ! center node 
    N = N + 1
    TAVAL(N) = APO - (AE + AW)
    TI(N) = NNT(I)
    TJ(N) = NNT(I)


    ! Neighbour nodes in domain
    IF(I.LT.IM1) THEN !East Node
      N = N + 1
      TAVAL(N) = AE
      Ti(N) = NNT(I)
      Tj(N) = NNT(I+1)

    ENDIF
    IF(I.GT.2) THEN !West Node
      N = N + 1
      TAVAL(N) = AW
      Ti(N) = NNT(I)
      Tj(N) = NNT(I-1)
    ENDIF
      
    TEMP_VEC(NNT(I))=TEMP(I)

  ENDDO    


RETURN
END

!------------------------------------------------------------
SUBROUTINE GEN_TB_VEC
!     purpose:
!     To construct b in Ax=b
USE FIELD
IMPLICIT NONE

DOUBLE PRECISION :: AE,AW


TB(:)=0.0

DO I=2,IM1

    TB(NNT(I))= TEMP_P(I)

    !At Boundary walls, T=known
    IF (I.EQ.IM1) THEN !East node
      AE = -DELT/DELX**2
      TB(NNT(I)) = TB(NNT(I)) - AE*TEMP_P(I+1)

    ENDIF
    IF (I.EQ.2) THEN !West node
      AW = -DELT/DELX**2
      TB(NNT(I)) = TB(NNT(I)) - AW*TEMP_P(I-1)
    ENDIF



ENDDO


RETURN
END

!------------------------------------------------------------
SUBROUTINE CORR_FS
!     purpose:
!     Correct solide fraction
USE FIELD
USE SOLVEFIELD
IMPLICIT NONE

DOUBLE PRECISION :: VAL1, VAL2, F_INV,FLAGVAL


FLAGVAL=1.0
! Initial guesses for Xi
DO I=2,IM1
    XI(I) = FS(I) - FS_P(I) - STE*(F_INV(FS(I),FSE,TS,TL) - TEMP(I))
    XI_N(I) = FS_N(I) - FS_P(I) - STE*(F_INV(FS_N(I),FSE,TS,TL) - TEMP(I))
ENDDO

! Secant Method
DO ITER_TT=1,ITMAX_TT
  ! Reset Residual
  RES_TT=0.0

  ! Loop pointwise through domain
  DO I=2,IM1

    ! Hold values for next iteration
    VAL1=FS(I)
    VAL2=XI(I)
      
    ! next value
    FS(I) = FS(I) - XI(I)*(FS(I)-FS_N(I))/(XI(I)-XI_N(I))
    XI(I) = FS(I) - FS_P(I) - STE*(F_INV(FS(I),FSE,TS,TL) - TEMP(I))

    ! Replace old values
    FS_N(I) = VAL1
    XI_N(I) = VAL2

    ! residual
    RES_TT = RES_TT + ABS(XI(I)-XI_N(I))
      
    ! flag value
    FLAGVAL = MIN(ABS(XI(I)-XI_N(I)), FLAGVAL)
  ENDDO


  ! Convergence test
  IF (RES_TT.LT.EPSI_TT) THEN
    EXIT
  ENDIF

  !flag val test
  IF (FLAGVAL.LT.(1E-15)) THEN
    EXIT
  ENDIF

ENDDO


! Limit Solid Fraction
DO I=2,IM1
  FS(I) = MIN( 1.0 , MAX( 0.0 , FS(I) ) )
ENDDO



RETURN
END


!------------------------------------------------------------
SUBROUTINE TITER
!     Purpose: Calculate temperature field
!     at current time level
USE FIELD
USE SOLVEFIELD
IMPLICIT NONE


! Store previous time step thermal fields
TEMP_P=TEMP
FS_P=FS

! Second starting guess of Secant method
FS_N=FS-0.1


  
! Generate b-Vector
CALL GEN_TB_VEC

! Calculate tentative Temperature field
CALL CGSOLVER(TAVAL,TI,TJ,TB,TEMP_VEC,UCOUNT_T,ENTRIESA_T,ITER_T,RES_T,EPSI_T,ITMAX_T)

! Update tentative temperature field
DO I=2,IM1
  TEMP(I) = TEMP_VEC(NNT(I))
ENDDO

! Solid fraction Correction
CALL CORR_FS

! Correct Temperature field
DO I=2,IM1
  TEMP(I) = TEMP(I) + (FS(I) - FS_P(I))/STE
ENDDO

! Impose BC
CALL BCTEMP

 



  
RETURN
END


!------------------------------------------------------------------------------------






! ========= CONJUGATE GRADIENT SOLVER =========


SUBROUTINE CGSOLVER(Aval,Ai,Aj,b,x,unknowns,entriesA, iter, res, epsi,itmax)	! Solves the system Ax=b with the conjugate gradients method

  IMPLICIT NONE
  INTEGER :: unknowns, entriesA,itmax,iter
  DOUBLE PRECISION, DIMENSION (entriesA) :: Aval
  INTEGER, DIMENSION (entriesA) :: Ai, Aj
  DOUBLE PRECISION, DIMENSION (unknowns) :: tempvec, x, r, d, b
                               
  DOUBLE PRECISION :: alpha, beta, tempscal1, tempscal2,epsi,res
  
  iter=0

  x(:) = 0.0
  r(:) = b(:)
  d(:) = r(:) 
  res = sum(abs(r))
  IF(res.LT.epsi) THEN		! Convergence test
     RETURN
  ENDIF


  DO iter=1,itmax
    CALL VECVEC(r,r,tempscal1,unknowns)			! (r'*r)
    CALL MATVEC(Aval,Ai,Aj,d,tempvec,unknowns,entriesA)	! (A*d)
    CALL VECVEC(d,tempvec,tempscal2,unknowns)		! (d'*A*d)
    alpha = tempscal1/tempscal2				! alpha = (r'*r)/(d'*A*d)
    x(:) = x(:) + alpha*d(:)				! x = x + alpha*d
    tempvec(:) = r(:) - alpha*tempvec(:)   		! r_new = r - alpha*A*d
    CALL VECVEC(tempvec,tempvec,tempscal2,unknowns)	! (r_new'*r_new)
    beta = tempscal2/tempscal1				! beta = (r_new'*r_new)/(r'*r)
    r(:) = tempvec(:)					! r = r_new
    d(:) = r(:) + beta*d(:)  				! d = r + beta*d

    res = sum(abs(r))
    IF(res.LT.epsi) THEN		! Convergence test
      EXIT
    ENDIF

  ENDDO
  

RETURN
END


!------------------------------------------------------------------------------------



SUBROUTINE MATVEC(Aval,Ai,Aj,u,x,m,n)	! Matrix-vector multiplication, x=Au  -- A: matrix, u: vector, x: return vector, m: length of u, n: entries in A
  IMPLICIT NONE
  INTEGER :: i, m, n
  DOUBLE PRECISION, DIMENSION(n) :: Aval 
  INTEGER, DIMENSION(n) :: Ai, Aj 
  DOUBLE PRECISION, DIMENSION(m) :: u,x
  
  x(:) = 0
  DO i=1,n
    x(Ai(i)) = x(Ai(i)) + Aval(i)*u(Aj(i))
  ENDDO
  
RETURN
END

SUBROUTINE MATTVEC(Aval,Ai,Aj,u,x,m,n)	! Transposed matrix-vector multiplication, x=Au  -- A: matrix, u: vector, x: return vector, m: length of u, n: entries in A
  IMPLICIT NONE
  INTEGER :: i, m, n
  DOUBLE PRECISION, DIMENSION(n) :: Aval 
  INTEGER, DIMENSION(n) :: Ai, Aj 
  DOUBLE PRECISION, DIMENSION(m) :: u,x
  
  x(:) = 0
  DO i=1,n
    x(Aj(i)) = x(Aj(i)) + Aval(i)*u(Ai(i))
  ENDDO
  
RETURN
END


SUBROUTINE VECVEC(u,v,x,m)	! Vector-vector multiplication, x=uv -- u: vector1, v: vector2, x: return scalar, m: size of u and v
  IMPLICIT NONE
  INTEGER :: i, m
  DOUBLE PRECISION, DIMENSION(m) :: u, v
  DOUBLE PRECISION :: x
  
  x = 0.0
  DO i=1,m
    x = x + u(i)*v(i)
  ENDDO
  
RETURN
END













